from evdev import UInput, AbsInfo, ecodes, events

from .types import AbsPosition


EVENTS_XBOX2EVDEV = {
    'a':             (events.EV_KEY, ecodes.BTN_A),
    'b':             (events.EV_KEY, ecodes.BTN_B),
    'x':             (events.EV_KEY, ecodes.BTN_X),
    'y':             (events.EV_KEY, ecodes.BTN_Y),
    'back':          (events.EV_KEY, ecodes.BTN_SELECT),
    'start':         (events.EV_KEY, ecodes.BTN_START),
    'leftshoulder':  (events.EV_KEY, ecodes.BTN_TL),
    'rightshoulder': (events.EV_KEY, ecodes.BTN_TR),
    'leftx':         (events.EV_ABS, ecodes.ABS_X),
    'lefty':         (events.EV_ABS, ecodes.ABS_Y),
    'rightx':        (events.EV_ABS, ecodes.ABS_RX),
    'righty':        (events.EV_ABS, ecodes.ABS_RY),
    'lefttrigger':   (events.EV_ABS, ecodes.ABS_Z),
    'righttrigger':  (events.EV_ABS, ecodes.ABS_RZ),
}


def xbox_event(controllerdb_key):
    """Return the evdev event equivalent of the Xbox controllerdb event

    Argument:
    controllerdb_key -- string (button a, b, x, y, ... or axes)
    Return:
    - a couple (event_type, event_code)
    - or None if key is unknown
    """
    if controllerdb_key in EVENTS_XBOX2EVDEV:
        return EVENTS_XBOX2EVDEV[controllerdb_key]
    else:
        return None


class Ubox360:
    """User space Gamepad (virtual device)"""

    def __init__(self):
        supported_events = {
            ecodes.EV_KEY: [
                ecodes.BTN_A,
                ecodes.BTN_B,
                ecodes.BTN_X,
                ecodes.BTN_Y,
                ecodes.BTN_TL,
                ecodes.BTN_TR,
                ecodes.BTN_SELECT,
                ecodes.BTN_START,
                ecodes.BTN_MODE,
                ecodes.BTN_THUMBL,
                ecodes.BTN_THUMBR],
            ecodes.EV_ABS: [
                (ecodes.ABS_X,     AbsInfo(value=0, min=0, max=255,
                                           fuzz=0, flat=15, resolution=0)),
                (ecodes.ABS_Y,     AbsInfo(value=0, min=0, max=255,
                                           fuzz=0, flat=15, resolution=0)),
                (ecodes.ABS_Z,     AbsInfo(value=0, min=0, max=1,
                                           fuzz=0, flat=0,  resolution=0)),
                (ecodes.ABS_RX,    AbsInfo(value=0, min=0, max=255,
                                           fuzz=0, flat=15, resolution=0)),
                (ecodes.ABS_RY,    AbsInfo(value=0, min=0, max=255,
                                           fuzz=0, flat=15, resolution=0)),
                (ecodes.ABS_RZ,    AbsInfo(value=0, min=0, max=1,
                                           fuzz=0, flat=0,  resolution=0)),
                (ecodes.ABS_HAT0X, AbsInfo(value=0, min=0, max=255,
                                           fuzz=0, flat=15, resolution=0)),
                (ecodes.ABS_HAT0Y, AbsInfo(value=0, min=0, max=255,
                                           fuzz=0, flat=15, resolution=0))]
        }
        self._udevice = UInput(supported_events, name="ubox360")

    def device_string(self):
        return str(self._udevice.device)

    def write_event(self, event_type, event_code, event_value):
        self._udevice.write(event_type, event_code, event_value)
        self._udevice.syn()

    def write_abs(self, event_type, event_code, abs_position):
        abs_info_list = self._udevice.device.capabilities()[event_type]
        abs_info = None
        for a in abs_info_list:
            if a[0] == event_code:
                abs_info = a[1]
        if abs_info is None:
            return
        if abs_position == AbsPosition.MIN:
            self._udevice.write(event_type, event_code, abs_info.min)
        elif abs_position == AbsPosition.MAX:
            self._udevice.write(event_type, event_code, abs_info.max)
        elif abs_position == AbsPosition.CENTER:
            event_value = (abs_info.max - abs_info.min) // 2
            self._udevice.write(event_type, event_code, event_value)
        self._udevice.syn()



__author__ = "Albin Kauffmann"
__email__ = "albin@kauff.org"
__description__ = "A tool to map joystick events to the one of an Xbox 360 controller"
__url__ = "https://gitlab.com/albinou/python-ubox360"
__license__ = "GPLv3"
__version__ = "0.3.0"



import evdev
import re

from .gamepad import Gamepad, gamepad_event
from .types import AbsPosition
from .ubox360 import Ubox360, xbox_event


def swap16b(val):
    return (((val & 0x00ff) << 8) |
            ((val & 0xff00) >> 8))


def parse_gamecontrollerdb(gamecontrollerdb_file):
    """Parse the gamecontrollerdb.txt file

    Argument:
    gamecontrollerdb_file -- path of the gamecontrollerdb.txt file to load
    Return:
    dict = [
      (vendor, product): [
        (control1, binding1),
        (control2, binding2),
        ...
        ]
      ...
    ]
    """
    controllerdb = {}
    pattern = re.compile(r"""                # Gamepad GUID:
                             ([0-9a-f]{4})   # - Device bus type (swapped)
                             0000
                             ([0-9a-f]{4})   # - Device vendor (swapped)
                             0000
                             ([0-9a-f]{4})   # - Device product (swapped)
                             0000
                             ([0-9a-f]{4})   # - version
                             0000,
                             ([^,]*),        # Gamepad name
                             (.*),           # key mappings + platform
                          """,
                         re.X)
    f = open(gamecontrollerdb_file)
    line = f.readline()
    while line:
        line = line.strip()
        m = pattern.fullmatch(line)
        if m:
            bus_type = swap16b(int(m.group(1), 16))  # noqa: F841
            vendor = swap16b(int(m.group(2), 16))
            product = swap16b(int(m.group(3), 16))
            version = swap16b(int(m.group(4), 16))   # noqa: F841
            gamepad_name = m.group(5)                # noqa: F841
            bindings = []
            for it in m.group(6).split(","):
                bindings.append(it.split(":"))
            controllerdb[(vendor, product)] = bindings
        line = f.readline()
    f.close()
    return controllerdb


def print_gamepad_addition(gamepad, ubox):
    print("Info: Adding a new gamepad mapping:")
    gamepad_strings = gamepad.device_strings()
    if len(gamepad_strings) > 1:
        size = len(gamepad_strings)
        end = size - 1
        half = end // 2
        for i in range(0, size):
            g = gamepad_strings[i]
            line_start = list("  | ")
            if i == 0:
                line_start[2:4] = ",-"
            elif i == end:
                line_start[2:4] = "'-"
            if i == half:
                if size % 2 == 0:
                    line_start[0:2] = " _"
                else:
                    line_start[0:2] = ",-"
            elif i > half:
                line_start[0] = "|"
            print("  %s%s" % ("".join(line_start), g))
    else:
        print("  ,-- %s" % (gamepad_strings[0]))
    print("  '-> %s" % (ubox.device_string()))


def create_devices(gamecontrollerdb, devices_path=None):
    """Return a list of (Gamepad, Ubox360) couples

    Argument:
    gamecontrollerdb -- result parse_gamecontrollerdb()
    devices_path -- if not None, only use devices from this list
    Return:
    list = [
      (Gamepad, Ubox360),
      ...
    ]
    """
    gamepads_list = []
    if devices_path is None:
        devices_path = evdev.list_devices()
    input_devices = [evdev.InputDevice(path) for path in devices_path]
    phys_devices = {}

    for d in input_devices:
        if (d.info.vendor, d.info.product) in gamecontrollerdb.keys():
            key = d.phys
            if "/" in key:
                key = key.split("/")[0]
            if key in phys_devices.keys():
                phys_devices[key].append(d)
            else:
                phys_devices[key] = [d]

    for gamepad_devices in phys_devices.values():
        d = gamepad_devices[0]
        evdev_mappings = {}
        for (key, value) in gamecontrollerdb[(d.info.vendor, d.info.product)]:
            if (key == "platform"):
                continue
            x_event = xbox_event(key)
            if (x_event is None):
                print("Warning: %s is an unknown Xbox event" % key)
                continue
            g_event = gamepad_event(d.capabilities(), value)
            if (g_event is None):
                print("Warning: %s is an unknown Gamepad event" % value)
                continue
            if all((type(e) == tuple) for e in g_event):
                evdev_mappings[g_event[0]] = x_event + (AbsPosition.MIN,)
                evdev_mappings[g_event[1]] = x_event + (AbsPosition.MAX,)
            else:
                evdev_mappings[g_event] = x_event
        g = Gamepad(gamepad_devices, evdev_mappings)
        u = Ubox360()
        print_gamepad_addition(g, u)
        gamepads_list.append((g, u))
    return gamepads_list



import argparse
import asyncio
import sys

from . import parse_gamecontrollerdb, create_devices
from .__about__ import (
    __description__,
    __version__,
)


async def main_loop(devices):
    tasks = []
    for (gamepad, ubox360) in devices:
        tasks += gamepad.create_asyncio_tasks(ubox360.write_event,
                                              ubox360.write_abs)
    for t in tasks:
        await t


def main():
    # parse args
    arg_parser = argparse.ArgumentParser(description=__description__)
    arg_parser.add_argument("-c",
                            "--controllerdb",
                            dest="controllerdb_file",
                            default="gamecontrollerdb.txt",
                            help="use this file gamecontroller database file")
    arg_parser.add_argument("-d",
                            "--device",
                            action="append",
                            dest="devices_path",
                            help="read inputs from this/these device(s)")
    arg_parser.add_argument("-v",
                            "--version",
                            action="version",
                            version="%s %s" % (__package__, __version__))
    args = arg_parser.parse_args()

    controllerdb = parse_gamecontrollerdb(args.controllerdb_file)
    devices = create_devices(controllerdb, args.devices_path)
    asyncio.run(main_loop(devices))
    return 0


if __name__ == '__main__':
    sys.exit(main())



import asyncio
import re

from evdev import ecodes, events

from .types import AbsPosition


class Gamepad:
    """Evdev Gamepad (real device)"""

    def __init__(self, devices, evdev_mappings):
        """Init the gamepad with events mapping facility

        Argument:
        devices -- list of InputDevice object (a game pad can have several
                   InputDevice object associated to it)
        evdev_mapping -- dictionnary of events (keys are real events and values
          are mapped events):
          {
            (event_type, event_code): (event_type, event_code, [AbsPosition]),
            ...
          }
        """
        self._devices = devices
        self._evdev_mappings = evdev_mappings

    def device_strings(self):
        """Return a list of devices strings"""
        str_list = []
        for d in self._devices:
            str_list.append(str(d))
        return str_list

    async def _read_events(self, device, write_xbox360_event,
                           write_xbox360_abs):
        async for event in device.async_read_loop():
            if (event.type, event.code) in self._evdev_mappings:
                e = self._evdev_mappings[(event.type, event.code)]
                if len(e) == 3:
                    if event.value == 0:
                        write_xbox360_abs(e[0], e[1], AbsPosition.CENTER)
                    else:
                        write_xbox360_abs(e[0], e[1], e[2])
                else:
                    write_xbox360_event(e[0], e[1], event.value)

    def create_asyncio_tasks(self, write_xbox360_event, write_xbox360_abs):
        tasks = []
        for d in self._devices:
            tasks.append(
                asyncio.create_task(
                    self._read_events(d, write_xbox360_event,
                                      write_xbox360_abs)
                )
            )
        return tasks


def gamepad_event(device_capabilities, controllerdb_key):
    """Return the evdev event equivalent of the real gamepad controllerdb event

    Argument:
    device_capabilities -- return of the InputDevice.capabilities() method
    controllerdb_key -- string (button b0, b1, b2, b3, ... or axes)
    Return one of these values:
    - a (event_type, event_code) for simple mapping
    - a ((event_type, event_code), (event_type, event_code)) tuple to represent
      an axis (obtained from 2 keys)
    - None if key is unknown
    """
    m = re.fullmatch(r"([ab])(\d+)|(\w+)(-(\w+))?", controllerdb_key)
    if m is None:
        return None
    elif m.group(1) == 'a':
        return (events.EV_ABS,
                device_capabilities[events.EV_ABS][int(m.group(2))][0])
    elif m.group(1) == 'b':
        return (events.EV_KEY,
                device_capabilities[events.EV_KEY][int(m.group(2))])
    elif (m.group(3) in ecodes.ecodes.keys() and
          m.group(5) is None):
        # Support keyboard mapping (non standard):
        return (events.EV_KEY, ecodes.ecodes[m.group(3)])
    elif (m.group(3) in ecodes.ecodes.keys() and
          m.group(5) in ecodes.ecodes.keys()):
        # Support axis mapping (non standard):
        return ((events.EV_KEY, ecodes.ecodes[m.group(3)]),
                (events.EV_KEY, ecodes.ecodes[m.group(5)]))
    else:
        return None
